import time
import uuid

import pytest
from dagster_graphql.test.utils import define_context_for_repository_yaml, execute_dagster_graphql

from dagster import seven
from dagster.core.instance import DagsterInstance, InstanceType
from dagster.core.scheduler import reconcile_scheduler_state
from dagster.core.scheduler.scheduler import ScheduleTickStatus
from dagster.core.storage.event_log import InMemoryEventLogStorage
from dagster.core.storage.local_compute_log_manager import NoOpComputeLogManager
from dagster.core.storage.root import LocalArtifactStorage
from dagster.core.storage.runs import InMemoryRunStorage
from dagster.core.storage.schedules.sqlite import SqliteScheduleStorage
from dagster.utils import file_relative_path
from dagster.utils.test import FilesytemTestScheduler

from .execution_queries import START_SCHEDULED_EXECUTION_QUERY
from .utils import InMemoryRunLauncher, sync_get_all_logs_for_run

SCHEDULE_TICKS_QUERY = '''
{
    scheduler {
    ... on Scheduler {
        runningSchedules {
            scheduleDefinition {
                name
            }
            ticks {
                tickId
                status
            }
            stats {
                ticksStarted
                ticksSucceeded
                ticksSkipped
                ticksFailed
            }
            ticksCount
        }
    }
    }
}
'''


def get_instance(temp_dir):
    return DagsterInstance(
        instance_type=InstanceType.EPHEMERAL,
        local_artifact_storage=LocalArtifactStorage(temp_dir),
        run_storage=InMemoryRunStorage(),
        event_storage=InMemoryEventLogStorage(),
        scheduler=FilesytemTestScheduler(temp_dir),
        schedule_storage=SqliteScheduleStorage.from_local(temp_dir),
        compute_log_manager=NoOpComputeLogManager(temp_dir),
    )


def get_instance_with_launcher(temp_dir):
    test_queue = InMemoryRunLauncher()

    return DagsterInstance(
        instance_type=InstanceType.EPHEMERAL,
        local_artifact_storage=LocalArtifactStorage(temp_dir),
        run_storage=InMemoryRunStorage(),
        event_storage=InMemoryEventLogStorage(),
        schedule_storage=SqliteScheduleStorage.from_local(temp_dir),
        compute_log_manager=NoOpComputeLogManager(temp_dir),
        run_launcher=test_queue,
    )


def test_basic_start_scheduled_execution():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'no_config_pipeline_hourly_schedule'},
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'

        assert uuid.UUID(result.data['startScheduledExecution']['run']['runId'])
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name']
            == 'no_config_pipeline'
        )

        assert any(
            tag['key'] == 'dagster/schedule_name'
            and tag['value'] == 'no_config_pipeline_hourly_schedule'
            for tag in result.data['startScheduledExecution']['run']['tags']
        )


def test_basic_start_scheduled_execution_with_run_launcher():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance_with_launcher(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'no_config_pipeline_hourly_schedule'},
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['startScheduledExecution']['__typename'] == 'LaunchPipelineRunSuccess'

        assert uuid.UUID(result.data['startScheduledExecution']['run']['runId'])
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name']
            == 'no_config_pipeline'
        )

        assert any(
            tag['key'] == 'dagster/schedule_name'
            and tag['value'] == 'no_config_pipeline_hourly_schedule'
            for tag in result.data['startScheduledExecution']['run']['tags']
        )


def test_basic_start_scheduled_execution_with_environment_dict_fn():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'no_config_pipeline_hourly_schedule_with_config_fn'},
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'

        assert uuid.UUID(result.data['startScheduledExecution']['run']['runId'])
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name']
            == 'no_config_pipeline'
        )

        assert any(
            tag['key'] == 'dagster/schedule_name'
            and tag['value'] == 'no_config_pipeline_hourly_schedule_with_config_fn'
            for tag in result.data['startScheduledExecution']['run']['tags']
        )


def test_start_scheduled_execution_with_should_execute():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'no_config_should_execute'},
        )

        assert not result.errors
        assert result.data

        assert result.data['startScheduledExecution']['__typename'] == 'ScheduledExecutionBlocked'


def test_partition_based_execution():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context, START_SCHEDULED_EXECUTION_QUERY, variables={'scheduleName': 'partition_based'},
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'

        assert uuid.UUID(result.data['startScheduledExecution']['run']['runId'])
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name']
            == 'no_config_pipeline'
        )

        tags = result.data['startScheduledExecution']['run']['tags']

        assert any(
            tag['key'] == 'dagster/schedule_name' and tag['value'] == 'partition_based'
            for tag in tags
        )

        assert any(tag['key'] == 'dagster/partition' and tag['value'] == '9' for tag in tags)
        assert any(
            tag['key'] == 'dagster/partition_set' and tag['value'] == 'scheduled_integer_partitions'
            for tag in tags
        )

        result_two = execute_dagster_graphql(
            context, START_SCHEDULED_EXECUTION_QUERY, variables={'scheduleName': 'partition_based'},
        )
        tags = result_two.data['startScheduledExecution']['run']['tags']
        # the last partition is selected on subsequent runs
        assert any(tag['key'] == 'dagster/partition' and tag['value'] == '9' for tag in tags)


def test_partition_based_custom_selector():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'partition_based_custom_selector'},
        )

        assert not result.errors
        assert result.data
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'
        assert uuid.UUID(result.data['startScheduledExecution']['run']['runId'])
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name']
            == 'no_config_pipeline'
        )
        tags = result.data['startScheduledExecution']['run']['tags']
        assert any(
            tag['key'] == 'dagster/schedule_name'
            and tag['value'] == 'partition_based_custom_selector'
            for tag in tags
        )
        assert any(tag['key'] == 'dagster/partition' and tag['value'] == '9' for tag in tags)
        assert any(
            tag['key'] == 'dagster/partition_set' and tag['value'] == 'scheduled_integer_partitions'
            for tag in tags
        )

        result_two = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'partition_based_custom_selector'},
        )
        tags = result_two.data['startScheduledExecution']['run']['tags']
        # get a different partition based on the subsequent run storage
        assert any(tag['key'] == 'dagster/partition' and tag['value'] == '8' for tag in tags)


def test_partition_based_decorator():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'partition_based_decorator'},
        )

        assert not result.errors
        assert result.data
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'


@pytest.mark.parametrize(
    'schedule_name',
    [
        'solid_subset_hourly_decorator',
        'solid_subset_daily_decorator',
        'solid_subset_monthly_decorator',
        'solid_subset_weekly_decorator',
    ],
)
def test_solid_subset_schedule_decorator(schedule_name):
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context, START_SCHEDULED_EXECUTION_QUERY, variables={'scheduleName': schedule_name},
        )

        assert not result.errors
        assert result.data
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'

        run_id = result.data['startScheduledExecution']['run']['runId']

        logs = sync_get_all_logs_for_run(context, run_id)['pipelineRunLogs']['messages']
        execution_step_names = [
            log['step']['key'] for log in logs if log['__typename'] == 'ExecutionStepStartEvent'
        ]
        assert execution_step_names == ['return_foo.compute']


def test_partition_based_multi_mode_decorator():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'partition_based_multi_mode_decorator'},
        )

        assert not result.errors
        assert result.data
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'
        run_id = result.data['startScheduledExecution']['run']['runId']

        logs = sync_get_all_logs_for_run(context, run_id)['pipelineRunLogs']['messages']
        execution_step_names = [
            log['step']['key'] for log in logs if log['__typename'] == 'ExecutionStepStartEvent'
        ]
        assert execution_step_names == ['return_six.compute']


# Tests for ticks and execution user error boundary
def test_tick_success(snapshot):
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )
        repository = context.get_repository()

        reconcile_scheduler_state("", "", repository, instance)
        schedule_def = repository.get_schedule_def("no_config_pipeline_hourly_schedule")

        start_time = time.time()
        execute_dagster_graphql(
            context, START_SCHEDULED_EXECUTION_QUERY, variables={'scheduleName': schedule_def.name},
        )

        # Check tick data and stats through gql
        result = execute_dagster_graphql(context, SCHEDULE_TICKS_QUERY)
        schedule_result = next(
            schedule_result
            for schedule_result in result.data['scheduler']['runningSchedules']
            if schedule_result['scheduleDefinition']['name'] == schedule_def.name
        )

        assert schedule_result
        assert schedule_result['stats']['ticksSucceeded'] == 1
        snapshot.assert_match(schedule_result)

        # Check directly against the DB
        ticks = instance.get_schedule_ticks_by_schedule(repository, schedule_def.name)
        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.schedule_name == schedule_def.name
        assert tick.cron_schedule == schedule_def.cron_schedule
        assert tick.timestamp > start_time and tick.timestamp < time.time()
        assert tick.status == ScheduleTickStatus.SUCCESS
        assert tick.run_id


def test_tick_skip(snapshot):
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )
        repository = context.get_repository()
        reconcile_scheduler_state("", "", repository, instance)

        execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'no_config_should_execute'},
        )

        # Check tick data and stats through gql
        result = execute_dagster_graphql(context, SCHEDULE_TICKS_QUERY)
        schedule_result = next(
            x
            for x in result.data['scheduler']['runningSchedules']
            if x['scheduleDefinition']['name'] == 'no_config_should_execute'
        )
        assert schedule_result['stats']['ticksSkipped'] == 1
        snapshot.assert_match(schedule_result)

        ticks = instance.get_schedule_ticks_by_schedule(repository, 'no_config_should_execute')

        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.status == ScheduleTickStatus.SKIPPED


def test_should_execute_scheduler_error(snapshot):
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )
        repository = context.get_repository()
        reconcile_scheduler_state("", "", repository, instance)

        execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'should_execute_error_schedule'},
        )

        # Check tick data and stats through gql
        result = execute_dagster_graphql(context, SCHEDULE_TICKS_QUERY)
        schedule_result = next(
            x
            for x in result.data['scheduler']['runningSchedules']
            if x['scheduleDefinition']['name'] == 'should_execute_error_schedule'
        )
        assert schedule_result['stats']['ticksFailed'] == 1
        snapshot.assert_match(schedule_result)

        ticks = instance.get_schedule_ticks_by_schedule(repository, 'should_execute_error_schedule')

        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.status == ScheduleTickStatus.FAILURE
        assert tick.error
        assert (
            "Error occurred during the execution should_execute for schedule "
            "should_execute_error_schedule" in tick.error.message
        )


def test_tags_scheduler_error(snapshot):
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )
        repository = context.get_repository()
        reconcile_scheduler_state("", "", repository, instance)

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'tags_error_schedule'},
        )
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'
        run_id = result.data['startScheduledExecution']['run']['runId']

        # Check tick data and stats through gql
        result = execute_dagster_graphql(context, SCHEDULE_TICKS_QUERY)
        schedule_result = next(
            x
            for x in result.data['scheduler']['runningSchedules']
            if x['scheduleDefinition']['name'] == 'tags_error_schedule'
        )

        assert schedule_result['stats']['ticksSucceeded'] == 1
        snapshot.assert_match(schedule_result)

        ticks = instance.get_schedule_ticks_by_schedule(repository, 'tags_error_schedule')

        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.status == ScheduleTickStatus.SUCCESS
        assert tick.run_id == run_id


def test_enviornment_dict_scheduler_error(snapshot):
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )
        repository = context.get_repository()
        reconcile_scheduler_state("", "", repository, instance)

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'environment_dict_error_schedule'},
        )
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'
        run_id = result.data['startScheduledExecution']['run']['runId']

        # Check tick data and stats through gql
        result = execute_dagster_graphql(context, SCHEDULE_TICKS_QUERY)
        schedule_result = next(
            x
            for x in result.data['scheduler']['runningSchedules']
            if x['scheduleDefinition']['name'] == 'environment_dict_error_schedule'
        )
        assert schedule_result['stats']['ticksSucceeded'] == 1
        snapshot.assert_match(schedule_result)

        ticks = instance.get_schedule_ticks_by_schedule(
            repository, 'environment_dict_error_schedule'
        )

        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.status == ScheduleTickStatus.SUCCESS
        assert tick.run_id == run_id


def test_enviornment_dict_scheduler_error_serialize_cauze():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )
        repository = context.get_repository()
        reconcile_scheduler_state("", "", repository, instance)

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'environment_dict_error_schedule'},
        )
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'
        run_id = result.data['startScheduledExecution']['run']['runId']

        ticks = instance.get_schedule_ticks_by_schedule(
            repository, 'environment_dict_error_schedule'
        )

        assert len(ticks) == 1
        tick = ticks[0]
        assert tick.status == ScheduleTickStatus.SUCCESS
        assert tick.run_id == run_id


def test_query_multiple_schedule_ticks(snapshot):
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )
        repository = context.get_repository()
        reconcile_scheduler_state("", "", repository, instance)

        for scheduleName in [
            'no_config_pipeline_hourly_schedule',
            'no_config_should_execute',
            'environment_dict_error_schedule',
        ]:
            execute_dagster_graphql(
                context, START_SCHEDULED_EXECUTION_QUERY, variables={'scheduleName': scheduleName},
            )

        result = execute_dagster_graphql(context, SCHEDULE_TICKS_QUERY)
        snapshot.assert_match(result.data['scheduler']['runningSchedules'])


def test_tagged_pipeline_schedule():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'tagged_pipeline_schedule'},
        )

        assert not result.errors
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name'] == 'tagged_pipeline'
        )

        assert any(
            tag['key'] == 'foo' and tag['value'] == 'bar'
            for tag in result.data['startScheduledExecution']['run']['tags']
        )


def test_tagged_pipeline_override_schedule():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'tagged_pipeline_override_schedule'},
        )

        assert not result.errors
        assert result.data['startScheduledExecution']['__typename'] == 'StartPipelineRunSuccess'
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name'] == 'tagged_pipeline'
        )

        assert not any(
            tag['key'] == 'foo' and tag['value'] == 'bar'
            for tag in result.data['startScheduledExecution']['run']['tags']
        )
        assert any(
            tag['key'] == 'foo' and tag['value'] == 'notbar'
            for tag in result.data['startScheduledExecution']['run']['tags']
        )


def test_tagged_pipeline_scheduled_execution_with_run_launcher():
    with seven.TemporaryDirectory() as temp_dir:
        instance = get_instance_with_launcher(temp_dir)
        context = define_context_for_repository_yaml(
            path=file_relative_path(__file__, '../repository.yaml'), instance=instance
        )

        result = execute_dagster_graphql(
            context,
            START_SCHEDULED_EXECUTION_QUERY,
            variables={'scheduleName': 'tagged_pipeline_schedule'},
        )

        assert not result.errors
        assert result.data

        # just test existence
        assert result.data['startScheduledExecution']['__typename'] == 'LaunchPipelineRunSuccess'

        assert uuid.UUID(result.data['startScheduledExecution']['run']['runId'])
        assert (
            result.data['startScheduledExecution']['run']['pipeline']['name'] == 'tagged_pipeline'
        )

        assert any(
            tag['key'] == 'foo' and tag['value'] == 'bar'
            for tag in result.data['startScheduledExecution']['run']['tags']
        )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_tags_scheduler_error 1'] = {
    'scheduleDefinition': {
        'name': 'tags_error_schedule'
    },
    'stats': {
        'ticksFailed': 0,
        'ticksSkipped': 0,
        'ticksStarted': 0,
        'ticksSucceeded': 1
    },
    'ticks': [
        {
            'status': 'SUCCESS',
            'tickId': '1'
        }
    ],
    'ticksCount': 1
}

snapshots['test_tick_skip 1'] = {
    'scheduleDefinition': {
        'name': 'no_config_should_execute'
    },
    'stats': {
        'ticksFailed': 0,
        'ticksSkipped': 1,
        'ticksStarted': 0,
        'ticksSucceeded': 0
    },
    'ticks': [
        {
            'status': 'SKIPPED',
            'tickId': '1'
        }
    ],
    'ticksCount': 1
}

snapshots['test_enviornment_dict_scheduler_error 1'] = {
    'scheduleDefinition': {
        'name': 'environment_dict_error_schedule'
    },
    'stats': {
        'ticksFailed': 0,
        'ticksSkipped': 0,
        'ticksStarted': 0,
        'ticksSucceeded': 1
    },
    'ticks': [
        {
            'status': 'SUCCESS',
            'tickId': '1'
        }
    ],
    'ticksCount': 1
}

snapshots['test_tick_success 1'] = {
    'scheduleDefinition': {
        'name': 'no_config_pipeline_hourly_schedule'
    },
    'stats': {
        'ticksFailed': 0,
        'ticksSkipped': 0,
        'ticksStarted': 0,
        'ticksSucceeded': 1
    },
    'ticks': [
        {
            'status': 'SUCCESS',
            'tickId': '1'
        }
    ],
    'ticksCount': 1
}

snapshots['test_should_execute_scheduler_error 1'] = {
    'scheduleDefinition': {
        'name': 'should_execute_error_schedule'
    },
    'stats': {
        'ticksFailed': 1,
        'ticksSkipped': 0,
        'ticksStarted': 0,
        'ticksSucceeded': 0
    },
    'ticks': [
        {
            'status': 'FAILURE',
            'tickId': '1'
        }
    ],
    'ticksCount': 1
}

snapshots['test_query_multiple_schedule_ticks 1'] = [
    {
        'scheduleDefinition': {
            'name': 'dynamic_config'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'environment_dict_error_schedule'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 1
        },
        'ticks': [
            {
                'status': 'SUCCESS',
                'tickId': '3'
            }
        ],
        'ticksCount': 1
    },
    {
        'scheduleDefinition': {
            'name': 'no_config_pipeline_hourly_schedule'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 1
        },
        'ticks': [
            {
                'status': 'SUCCESS',
                'tickId': '1'
            }
        ],
        'ticksCount': 1
    },
    {
        'scheduleDefinition': {
            'name': 'no_config_pipeline_hourly_schedule_with_config_fn'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'no_config_should_execute'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 1,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
            {
                'status': 'SKIPPED',
                'tickId': '2'
            }
        ],
        'ticksCount': 1
    },
    {
        'scheduleDefinition': {
            'name': 'partition_based'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'partition_based_custom_selector'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'partition_based_decorator'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'partition_based_multi_mode_decorator'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'should_execute_error_schedule'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'solid_subset_daily_decorator'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'solid_subset_hourly_decorator'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'solid_subset_monthly_decorator'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'solid_subset_weekly_decorator'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'tagged_pipeline_override_schedule'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'tagged_pipeline_schedule'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    },
    {
        'scheduleDefinition': {
            'name': 'tags_error_schedule'
        },
        'stats': {
            'ticksFailed': 0,
            'ticksSkipped': 0,
            'ticksStarted': 0,
            'ticksSucceeded': 0
        },
        'ticks': [
        ],
        'ticksCount': 0
    }
]



from dagster_graphql.implementation.utils import ExecutionParams

from dagster import PipelineDefinition, check
from dagster.core.execution.api import create_execution_plan
from dagster.core.execution.memoization import get_retry_steps_from_execution_plan
from dagster.core.instance import DagsterInstance
from dagster.core.storage.pipeline_run import PipelineRunStatus
from dagster.core.storage.tags import RESUME_RETRY_TAG
from dagster.core.utils import make_new_run_id


def pipeline_run_args_from_execution_params(execution_params, step_keys_to_execute=None):
    check.inst_param(execution_params, 'execution_params', ExecutionParams)
    return dict(
        pipeline_name=execution_params.selector.name,
        run_id=execution_params.execution_metadata.run_id
        if execution_params.execution_metadata.run_id
        else make_new_run_id(),
        solid_subset=execution_params.selector.solid_subset,
        environment_dict=execution_params.environment_dict,
        mode=execution_params.mode,
        step_keys_to_execute=step_keys_to_execute or execution_params.step_keys,
        tags=execution_params.execution_metadata.tags,
        root_run_id=execution_params.execution_metadata.root_run_id,
        parent_run_id=execution_params.execution_metadata.parent_run_id,
        status=PipelineRunStatus.NOT_STARTED,
    )


def get_step_keys_to_execute(instance, pipeline_def, execution_params):
    check.inst_param(instance, 'instance', DagsterInstance)
    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    check.inst_param(execution_params, 'execution_params', ExecutionParams)

    if not execution_params.step_keys and is_resume_retry(execution_params):
        # Get step keys from parent_run_id if it's a resume/retry
        execution_plan = create_execution_plan(
            pipeline_def, execution_params.environment_dict, mode=execution_params.mode,
        )
        return get_retry_steps_from_execution_plan(
            instance, execution_plan, execution_params.execution_metadata.parent_run_id
        )
    else:
        return execution_params.step_keys


def is_resume_retry(execution_params):
    check.inst_param(execution_params, 'execution_params', ExecutionParams)
    return execution_params.execution_metadata.tags.get(RESUME_RETRY_TAG) == 'true'



from dagster_graphql.test.utils import execute_dagster_graphql
from dagster_graphql_tests.graphql.setup import define_test_subprocess_context
from dagster_graphql_tests.test_cli import START_PIPELINE_EXECUTION_QUERY

from dagster.core.instance import DagsterInstance
from dagster.seven.abc import Iterable
from dagster.utils import file_relative_path


def is_clear_execution_manager(manager):
    manager_process_state = [
        getattr(manager, field)
        for field in vars(manager)
        if isinstance(getattr(manager, field), Iterable)
    ]
    # Ensure no memory growth over time in long living dagits
    assert all([len(iterable) == 0 for iterable in manager_process_state])


def test_term_event_lifecycle():
    context = define_test_subprocess_context(DagsterInstance.local_temp())
    execution_manager = context.execution_manager
    is_clear_execution_manager(execution_manager)
    result = execute_dagster_graphql(
        context,
        START_PIPELINE_EXECUTION_QUERY,
        variables={
            'executionParams': {
                'selector': {'name': 'csv_hello_world'},
                'mode': 'default',
                'environmentConfigData': {
                    'execution': {'multiprocess': {'config': {'max_concurrent': 4}}},
                    'solids': {
                        'sum_solid': {
                            'inputs': {'num': file_relative_path(__file__, '../data/num.csv')}
                        }
                    },
                },
            }
        },
    )

    # Make sure pipeline ran successfully
    assert not result.errors
    assert result.data
    assert result.data['startPipelineExecution']['__typename'] == 'StartPipelineRunSuccess'

    # Clean up
    execution_manager.join()
    execution_manager.check()
    is_clear_execution_manager(execution_manager)



import re

from dagster_graphql.test.utils import execute_dagster_graphql

from .setup import define_test_context

SCHEMA_OR_ERROR_SUBSET_QUERY = '''
query EnvironmentQuery($pipelineName: String!, $solidSubset: [String!]){
    environmentSchemaOrError(selector :{ name: $pipelineName, solidSubset: $solidSubset}) {
        __typename
        ... on EnvironmentSchema {
            allConfigTypes {
                __typename
                key
                ... on CompositeConfigType {
                    __typename
                    fields {
                        __typename
                        name
                        configType {
                            key
                            __typename
                        }
                    }
                }
            }
        }
        ... on InvalidSubsetError {
            message
        }
    }
}
'''


def field_names_of(type_dict, typename):
    return {field_data['name'] for field_data in type_dict[typename]['fields']}


def types_dict_of_result(subset_result, top_key):
    return {
        type_data['name']: type_data for type_data in subset_result.data[top_key]['configTypes']
    }


def test_csv_hello_world_pipeline_or_error_subset_wrong_solid_name():
    result = execute_dagster_graphql(
        define_test_context(),
        SCHEMA_OR_ERROR_SUBSET_QUERY,
        {'pipelineName': 'csv_hello_world', 'solidSubset': ['nope']},
    )

    assert not result.errors
    assert result.data
    assert result.data['environmentSchemaOrError']['__typename'] == 'InvalidSubsetError'
    assert '"nope" does not exist' in result.data['environmentSchemaOrError']['message']


def test_pipeline_with_invalid_definition_error():
    result = execute_dagster_graphql(
        define_test_context(),
        SCHEMA_OR_ERROR_SUBSET_QUERY,
        {'pipelineName': 'pipeline_with_invalid_definition_error', 'solidSubset': ['fail_subset'],},
    )
    assert not result.errors
    assert result.data
    assert result.data['environmentSchemaOrError']['__typename'] == 'InvalidSubsetError'
    assert re.match(
        (
            r'.*DagsterInvalidDefinitionError[\s\S]*'
            r'add a input_hydration_config for the type "InputTypeWithoutHydration"'
        ),
        result.data['environmentSchemaOrError']['message'],
    )



import sys

from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.config.validate import validate_config_from_snap
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from dagster.core.snap import (
    ActivePipelineData,
    ExecutionPlanIndex,
    PipelineIndex,
    active_pipeline_data_from_def,
)
from dagster.utils.error import serializable_error_info_from_exc_info

from .utils import UserFacingGraphQLError


class __SolidSubsetNotProvidedSentinel(object):
    pass


SOLID_SUBSET_NOT_PROVIDED = __SolidSubsetNotProvidedSentinel


# Represents a pipeline definition that is resident in an external process.
#
# Object composes a pipeline index (which is an index over snapshot data)
# and the serialized ActivePipelineData
class ExternalPipeline:
    def __init__(
        self, pipeline_index, active_pipeline_data, solid_subset=SOLID_SUBSET_NOT_PROVIDED,
    ):
        self.pipeline_index = check.inst_param(pipeline_index, 'pipeline_index', PipelineIndex)
        self._active_pipeline_data = check.inst_param(
            active_pipeline_data, 'active_pipeline_data', ActivePipelineData
        )
        self._active_preset_dict = {ap.name: ap for ap in active_pipeline_data.active_presets}

        if solid_subset != SOLID_SUBSET_NOT_PROVIDED:
            check.opt_list_param(solid_subset, 'solid_subset', str)

        self._solid_subset = solid_subset

    @property
    def name(self):
        return self.pipeline_index.name

    @property
    def solid_subset(self):
        if self._solid_subset == SOLID_SUBSET_NOT_PROVIDED:
            raise DagsterInvariantViolationError(
                "Cannot access property solid_subset on external pipeline constructed without a "
                "solid subset"
            )

        return self._solid_subset

    @property
    def active_presets(self):
        return self._active_pipeline_data.active_presets

    @property
    def pipeline_snapshot(self):
        return self.pipeline_index.pipeline_snapshot

    def has_solid_invocation(self, solid_name):
        check.str_param(solid_name, 'solid_name')
        return self.pipeline_index.has_solid_invocation(solid_name)

    def has_preset(self, preset_name):
        check.str_param(preset_name, 'preset_name')
        return preset_name in self._active_preset_dict

    def get_preset(self, preset_name):
        check.str_param(preset_name, 'preset_name')
        return self._active_preset_dict[preset_name]

    def get_mode(self, mode_name):
        check.str_param(mode_name, 'mode_name')
        return self.pipeline_index.get_mode_def_snap(mode_name)

    @staticmethod
    def from_pipeline_def(pipeline_def, solid_subset=None):
        if solid_subset:
            pipeline_def = pipeline_def.build_sub_pipeline(solid_subset)

        return ExternalPipeline(
            pipeline_def.get_pipeline_index(),
            active_pipeline_data_from_def(pipeline_def),
            solid_subset=solid_subset,
        )

    @property
    def config_schema_snapshot(self):
        return self.pipeline_index.pipeline_snapshot.config_schema_snapshot

    @property
    def pipeline_snapshot_id(self):
        return self.pipeline_index.pipeline_snapshot_id

    def get_default_mode_name(self):
        return self.pipeline_index.get_default_mode_name()

    @property
    def tags(self):
        return self.pipeline_index.pipeline_snapshot.tags


def get_external_pipeline_or_raise(graphene_info, pipeline_name):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)

    if not graphene_info.context.has_external_pipeline(pipeline_name):
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PipelineNotFoundError')(pipeline_name=pipeline_name)
        )

    return graphene_info.context.get_external_pipeline(pipeline_name)


def get_external_pipeline_subset_or_raise(graphene_info, pipeline_name, solid_subset):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.str_param(pipeline_name, 'pipeline_name')
    check.opt_list_param(solid_subset, 'solid_subset', of_type=str)

    from dagster_graphql.schema.errors import DauphinInvalidSubsetError

    full_pipeline = get_external_pipeline_or_raise(graphene_info, pipeline_name)

    if solid_subset is None:
        return full_pipeline

    for solid_name in solid_subset:
        if not full_pipeline.has_solid_invocation(solid_name):
            raise UserFacingGraphQLError(
                DauphinInvalidSubsetError(
                    message='Solid "{solid_name}" does not exist in "{pipeline_name}"'.format(
                        solid_name=solid_name, pipeline_name=pipeline_name
                    ),
                    pipeline=graphene_info.schema.type_named('Pipeline')(full_pipeline),
                )
            )
    try:
        return graphene_info.context.get_external_pipeline_subset(pipeline_name, solid_subset)
    except DagsterInvalidDefinitionError:
        # this handles the case when you construct a subset such that an unsatisfied
        # input cannot be hydrate from config. Current this is only relevant for
        # the in-process case. Once we add the out-of-process we will communicate
        # this error through the communication channel and change what exception
        # is thrown
        raise UserFacingGraphQLError(
            DauphinInvalidSubsetError(
                message=serializable_error_info_from_exc_info(sys.exc_info()).message,
                pipeline=graphene_info.schema.type_named('Pipeline')(full_pipeline),
            )
        )


def ensure_valid_config(external_pipeline, mode, environment_dict):
    check.inst_param(external_pipeline, 'external_pipeline', ExternalPipeline)
    check.str_param(mode, 'mode')
    # do not type check environment_dict so that validate_config_from_snap throws

    validated_config = validate_config_from_snap(
        config_schema_snapshot=external_pipeline.config_schema_snapshot,
        config_type_key=external_pipeline.pipeline_index.get_mode_def_snap(mode).root_config_key,
        config_value=environment_dict,
    )

    if not validated_config.success:
        from dagster_graphql.schema.errors import DauphinPipelineConfigValidationInvalid

        raise UserFacingGraphQLError(
            DauphinPipelineConfigValidationInvalid.for_validation_errors(
                external_pipeline.pipeline_index, validated_config.errors
            )
        )

    return validated_config


def ensure_valid_step_keys(full_execution_plan_index, step_keys):
    check.inst_param(full_execution_plan_index, 'full_execution_plan_index', ExecutionPlanIndex)
    check.opt_list_param(step_keys, 'step_keys', of_type=str)

    if not step_keys:
        return

    for step_key in step_keys:
        if not full_execution_plan_index.has_step(step_key):
            from dagster_graphql.schema.errors import DauphinInvalidStepError

            raise UserFacingGraphQLError(DauphinInvalidStepError(invalid_step_key=step_key))


def get_execution_plan_index_or_raise(
    graphene_info, external_pipeline, mode, environment_dict, step_keys_to_execute
):
    full_execution_plan_index = graphene_info.context.create_execution_plan_index(
        external_pipeline=external_pipeline, environment_dict=environment_dict, mode=mode,
    )

    if not step_keys_to_execute:
        return full_execution_plan_index

    ensure_valid_step_keys(full_execution_plan_index, step_keys_to_execute)

    return graphene_info.context.create_execution_plan_index(
        external_pipeline=external_pipeline,
        environment_dict=environment_dict,
        mode=mode,
        step_keys_to_execute=step_keys_to_execute,
    )



from __future__ import absolute_import

import sys
import time

from dagster_graphql.implementation.utils import ExecutionMetadata, ExecutionParams
from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.config.validate import validate_config_from_snap
from dagster.core.definitions.schedule import ScheduleExecutionContext
from dagster.core.errors import (
    DagsterUserCodeExecutionError,
    ScheduleExecutionError,
    user_code_error_boundary,
)
from dagster.core.events import EngineEventData
from dagster.core.scheduler import ScheduleTickStatus
from dagster.core.scheduler.scheduler import ScheduleTickData
from dagster.core.storage.tags import check_tags
from dagster.utils.error import serializable_error_info_from_exc_info
from dagster.utils.merger import merge_dicts

from ..external import ExternalPipeline, get_external_pipeline_subset_or_raise
from ..fetch_schedules import get_dagster_schedule_def
from ..utils import capture_dauphin_error
from .launch_execution import _launch_pipeline_execution_for_created_run
from .start_execution import _start_pipeline_execution_for_created_run


@capture_dauphin_error
def start_scheduled_execution(graphene_info, schedule_name):
    '''
    When a scheduler ticks and needs to run for a given schedule, it issues a
    START_SCHEDULED_EXECUTION mutation with just the schedule name. The mutation is
    resolved entirely by this method.
    '''

    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.str_param(schedule_name, 'schedule_name')

    tick = None
    try:
        # We first load the repository and schedule definition to create
        # and store a ScheduleTick.
        # If this fails, this error should be sent to the file based scheduler logs.
        repository = graphene_info.context.get_repository()
        schedule_def = get_dagster_schedule_def(graphene_info, schedule_name)
        cron_schedule = "Unknown" if not schedule_def else schedule_def.cron_schedule
        tick = graphene_info.context.instance.create_schedule_tick(
            repository,
            ScheduleTickData(
                schedule_name=schedule_name,
                cron_schedule=cron_schedule,
                timestamp=time.time(),
                status=ScheduleTickStatus.STARTED,
            ),
        )

        # Run should_execute and halt if it returns False
        schedule_context = ScheduleExecutionContext(graphene_info.context.instance, repository)
        with user_code_error_boundary(
            ScheduleExecutionError,
            lambda: 'Error occurred during the execution should_execute for schedule '
            '{schedule_name}'.format(schedule_name=schedule_def.name),
        ):
            should_execute = schedule_def.should_execute(schedule_context)

        if not should_execute:
            # Update tick to skipped state and return
            tick = tick.with_status(ScheduleTickStatus.SKIPPED)
            graphene_info.context.instance.update_schedule_tick(repository, tick)
            # Return skipped specific gql response
            return graphene_info.schema.type_named('ScheduledExecutionBlocked')(
                message='Schedule {schedule_name} did not run because the should_execute did not return'
                ' True'.format(schedule_name=schedule_name)
            )

        errors = []

        environment_dict = {}
        schedule_tags = {}
        try:
            with user_code_error_boundary(
                ScheduleExecutionError,
                lambda: 'Error occurred during the execution of environment_dict_fn for schedule '
                '{schedule_name}'.format(schedule_name=schedule_def.name),
            ):
                environment_dict = schedule_def.get_environment_dict(schedule_context)
        except DagsterUserCodeExecutionError as exc:
            error_data = serializable_error_info_from_exc_info(sys.exc_info())
            errors.append(error_data)

        try:
            with user_code_error_boundary(
                ScheduleExecutionError,
                lambda: 'Error occurred during the execution of tags_fn for schedule '
                '{schedule_name}'.format(schedule_name=schedule_def.name),
            ):
                schedule_tags = schedule_def.get_tags(schedule_context)
        except DagsterUserCodeExecutionError:
            error_data = serializable_error_info_from_exc_info(sys.exc_info())
            errors.append(error_data)

        external_pipeline = get_external_pipeline_subset_or_raise(
            graphene_info, schedule_def.selector.name, schedule_def.selector.solid_subset
        )
        pipeline_tags = external_pipeline.tags or {}
        check_tags(pipeline_tags, 'pipeline_tags')
        tags = merge_dicts(pipeline_tags, schedule_tags)

        selector = schedule_def.selector
        mode = schedule_def.mode

        execution_params = ExecutionParams(
            selector=selector,
            environment_dict=environment_dict,
            mode=mode,
            execution_metadata=ExecutionMetadata(tags=tags, run_id=None),
            step_keys=None,
        )

        run, result = _execute_schedule(graphene_info, external_pipeline, execution_params, errors)
        graphene_info.context.instance.update_schedule_tick(
            repository, tick.with_status(ScheduleTickStatus.SUCCESS, run_id=run.run_id),
        )

        return result

    except Exception as exc:  # pylint: disable=broad-except
        error_data = serializable_error_info_from_exc_info(sys.exc_info())

        if tick:
            graphene_info.context.instance.update_schedule_tick(
                repository, tick.with_status(ScheduleTickStatus.FAILURE, error=error_data),
            )

        raise exc


def _execute_schedule(graphene_info, external_pipeline, execution_params, errors):
    check.inst_param(external_pipeline, 'external_pipeline', ExternalPipeline)

    instance = graphene_info.context.instance

    mode, environment_dict = execution_params.mode, execution_params.environment_dict

    validation_result = validate_config_from_snap(
        external_pipeline.config_schema_snapshot,
        external_pipeline.get_mode(mode).root_config_key,
        environment_dict,
    )
    if validation_result.success:
        execution_plan_index = graphene_info.context.create_execution_plan_index(
            external_pipeline, environment_dict, mode, execution_params.step_keys
        )

    pipeline_run = instance.create_run(
        pipeline_name=external_pipeline.name,
        environment_dict=environment_dict,
        mode=mode,
        solid_subset=(
            execution_params.selector.solid_subset
            if execution_params.selector is not None
            else None
        ),
        tags=execution_params.execution_metadata.tags,
        pipeline_snapshot=external_pipeline.pipeline_snapshot,
        execution_plan_snapshot=execution_plan_index.execution_plan_snapshot,
    )

    # Inject errors into event log at this point
    if len(errors) > 0:
        for error in errors:
            instance.report_engine_event(
                error.message, pipeline_run, EngineEventData.engine_error(error)
            )

    # Launch run if run launcher is defined
    run_launcher = graphene_info.context.instance.run_launcher
    if run_launcher:
        result = _launch_pipeline_execution_for_created_run(graphene_info, pipeline_run.run_id)
    else:
        result = _start_pipeline_execution_for_created_run(graphene_info, pipeline_run.run_id)

    return pipeline_run, result



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots['test_pipeline_reexecution_info_query 1'] = [
    'sum_sq_solid.compute'
]



from dagster_graphql.test.utils import execute_dagster_graphql

from .setup import define_test_context, define_test_snapshot_context


def all_solids_query():
    return '{ usedSolids { __typename, definition { name }, invocations { pipeline { name }, solidHandle { handleID } } } }'


def get_solid_query_exists():
    return '''
    { 
        usedSolid(name: "sum_solid") {
            ... on UsedSolid { definition { name } }
        }
    }
    '''


def test_query_all_solids(snapshot):
    result = execute_dagster_graphql(define_test_context(), all_solids_query())
    snapshot.assert_match(result.data)


def test_query_all_solids_with_snapshot_context(snapshot):
    result = execute_dagster_graphql(define_test_snapshot_context(), all_solids_query(),)
    snapshot.assert_match(result.data)


def test_query_get_solid_exists():
    result = execute_dagster_graphql(define_test_context(), get_solid_query_exists())

    assert not result.errors
    assert result.data['usedSolid']['definition']['name'] == 'sum_solid'


def test_query_get_solid_with_snapshot_context_exists():
    result = execute_dagster_graphql(define_test_snapshot_context(), get_solid_query_exists(),)

    assert not result.errors
    assert result.data['usedSolid']['definition']['name'] == 'sum_solid'



from graphql.execution.base import ResolveInfo

from dagster import check
from dagster.config.validate import validate_config_from_snap
from dagster.core.definitions.pipeline import ExecutionSelector
from dagster.core.snap import PipelineIndex

from .external import get_external_pipeline_subset_or_raise
from .utils import UserFacingGraphQLError, capture_dauphin_error


@capture_dauphin_error
def resolve_environment_schema_or_error(graphene_info, selector, mode):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(selector, 'selector', ExecutionSelector)
    check.opt_str_param(mode, 'mode')

    pipeline_index = get_external_pipeline_subset_or_raise(
        graphene_info, selector.name, selector.solid_subset
    ).pipeline_index

    if mode is None:
        mode = pipeline_index.get_default_mode_name()

    if not pipeline_index.has_mode_def(mode):
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('ModeNotFoundError')(mode=mode, selector=selector)
        )

    return graphene_info.schema.type_named('EnvironmentSchema')(
        pipeline_index=pipeline_index, mode=mode,
    )


@capture_dauphin_error
def resolve_is_environment_config_valid(graphene_info, pipeline_index, mode, environment_dict):
    check.inst_param(graphene_info, 'graphene_info', ResolveInfo)
    check.inst_param(pipeline_index, 'pipeline_index', PipelineIndex)
    check.str_param(mode, 'mode')
    check.dict_param(environment_dict, 'environment_dict', key_type=str)

    mode_def_snap = pipeline_index.get_mode_def_snap(mode)

    if not mode_def_snap.root_config_key:
        # historical pipeline with unknown environment type. blindly pass validation
        return graphene_info.schema.type_named('PipelineConfigValidationValid')(pipeline_index.name)

    validated_config = validate_config_from_snap(
        pipeline_index.config_schema_snapshot, mode_def_snap.root_config_key, environment_dict
    )

    if not validated_config.success:
        raise UserFacingGraphQLError(
            graphene_info.schema.type_named('PipelineConfigValidationInvalid')(
                pipeline_name=pipeline_index.name,
                errors=[
                    graphene_info.schema.type_named(
                        'PipelineConfigValidationError'
                    ).from_dagster_error(
                        pipeline_index.config_schema_snapshot, err,
                    )
                    for err in validated_config.errors
                ],
            )
        )

    return graphene_info.schema.type_named('PipelineConfigValidationValid')(pipeline_index.name)



from dagster_graphql.test.utils import define_context_for_repository_yaml, execute_dagster_graphql

from dagster.utils import file_relative_path

GET_PARTITION_SETS_QUERY = '''
{
    partitionSetsOrError {
        __typename
        ...on PartitionSets {
            results {
                name
                pipelineName
                solidSubset
                mode
            }
        }
    }
}
'''

GET_PARTITION_SETS_FOR_PIPELINE_QUERY = '''
    query PartitionSetsQuery($pipelineName: String!) {
        partitionSetsOrError(pipelineName: $pipelineName) {
            __typename
            ...on PartitionSets {
                results {
                    name
                    pipelineName
                    solidSubset
                    mode
                }
            }
            ...on PipelineNotFoundError {
                message
            }
        }
    }
'''

GET_PARTITION_SET_QUERY = '''
    query ParitionSetQuery($partitionSetName: String!) {
        partitionSetOrError(partitionSetName: $partitionSetName) {
            __typename
            ...on PartitionSet {
                name
                pipelineName
                solidSubset
                mode
                partitions {
                    results {
                        name
                    }
                }
            }
        }
    }
'''


def test_get_all_partition_sets(snapshot):
    context = define_context_for_repository_yaml(
        path=file_relative_path(__file__, '../repository.yaml')
    )

    result = execute_dagster_graphql(context, GET_PARTITION_SETS_QUERY)

    assert result.data
    snapshot.assert_match(result.data)


def test_get_partition_sets_for_pipeline(snapshot):
    context = define_context_for_repository_yaml(
        path=file_relative_path(__file__, '../repository.yaml')
    )

    result = execute_dagster_graphql(
        context,
        GET_PARTITION_SETS_FOR_PIPELINE_QUERY,
        variables={'pipelineName': 'no_config_pipeline'},
    )

    assert result.data
    snapshot.assert_match(result.data)

    invalid_pipeline_result = execute_dagster_graphql(
        context,
        GET_PARTITION_SETS_FOR_PIPELINE_QUERY,
        variables={'pipelineName': 'invalid_pipeline'},
    )

    assert invalid_pipeline_result.data
    snapshot.assert_match(invalid_pipeline_result.data)


def test_get_partition_set(snapshot):
    context = define_context_for_repository_yaml(
        path=file_relative_path(__file__, '../repository.yaml')
    )

    result = execute_dagster_graphql(
        context, GET_PARTITION_SET_QUERY, variables={'partitionSetName': 'integer_partition'},
    )

    assert result.data
    snapshot.assert_match(result.data)

    invalid_partition_set_result = execute_dagster_graphql(
        context, GET_PARTITION_SET_QUERY, variables={'partitionSetName': 'invalid_partition'},
    )

    assert invalid_partition_set_result.data
    snapshot.assert_match(invalid_partition_set_result.data)
